@JS()
library web_components;

import 'dart:async';
import 'dart:html';

import 'package:js/js.dart';

bool _customElementsEnabled = false;

void enableCustomElements() {
  if (!_customElementsEnabled) {
    eval('''
      class CustomElement extends HTMLElement {
        constructor() {
          super();
          this.connected = null;
          this.disconnected = null;
        }
      
        connectedCallback() {
          this.connected();
        }
      
        disconnectedCallback() {
          this.disconnected();
        }
      }
      window.CustomElement = CustomElement;
      
      function nativeDefineElement(name, construct) {
        customElements.define(name, class extends CustomElement {
          constructor() {
            super();
            construct(this);
          }
        });
      }
      
      function nativeAddCallbacks(element, connected, disconnected) {
        element.connected = connected;
        element.disconnected = disconnected;
      }
      
      function nativeAttachShadow(element, mode, delegatesFocus) {
        return element.attachShadow({mode: mode, delegatesFocus: delegatesFocus});
      }
    ''');
    _customElementsEnabled = true;
  }
}

@JS('eval')
external void eval(String javascript);

typedef CustomElementConstructor = CustomElement Function(HtmlElement e);

typedef Callback = void Function();

void defineElement(String name, CustomElementConstructor constructor) {
  nativeDefineElement(name, allowInterop(constructor));
}

@JS('nativeDefineElement')
external void nativeDefineElement(String name, CustomElementConstructor constructor);

@JS('nativeAddCallbacks')
external void nativeAddCallbacks(HtmlElement element, Callback connected, Callback disconnected);

@JS('nativeAttachShadow')
external ShadowRoot nativeAttachShadow(HtmlElement element, String mode, bool delegatesFocus);

Map<HtmlElement, CustomElement> _components = {};

extension HtmlElementComponent on HtmlElement {
  set component(CustomElement value) => _components[this] = value;

  CustomElement get component => _components[this];
}

class CustomElement {
  static final List<CustomElement> _elementStack = [];
  final HtmlElement element;

  CustomElement(this.element) {
    nativeAddCallbacks(element, allowInterop(connectedCallback), allowInterop(disconnected));
    element.component = this;
  }

  void connectedCallback() {
    connected();
    _elementStack.add(this);
    Timer(Duration(seconds: 0), resolveChildren);
  }

  void resolveChildren() {
    while (_elementStack.isNotEmpty) {
      var element = _elementStack.removeLast();
      element.childrenConnected();
    }
  }

  void connected() {}

  void childrenConnected() {}

  void disconnected() {}

  ShadowRoot attachShadow({bool open = true, bool delegatesFocus = false}) =>
      nativeAttachShadow(element, open ? 'open' : 'closed', delegatesFocus);

  T parentOfType<T>() {
    var parent = element.parentNode;
    if (parent is HtmlElement) {
      if (parent.component is T) {
        return parent.component as T;
      }
    }
    throw 'No parent found of type $T';
  }
}
